//
// Hilligoss-HMSFC-512 v1.0 released 062524:0051 by Bus Error Collective
// 
// Hilligoss will convert any 8-bit 512x512 Binary/"Raw" PGM image to PCM stereo audio data
// suitable for layback on XY-mode oscilloscopes and similar visualizers. It is flexible 
// enough to accomodate a wide range of sample/frame rates, and smart enough about image
// quality to render excellent looking output. 
//
// Hilligoss uses an image reduction and conversion algorithm which favors bright pixels
// over dark pixels when rendering.  It also employs a Hilbert curve to determine the 
// closest path betwen neighboring points.
//
// Thanks to MrsBeanbag for her help on space-filling curves!
//
// Hilbert-Moore Space Filling Curve version


#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>
#include <limits.h>
#include <string.h>

FILE* inputFile;		// File handle..
FILE* outputFile;		// File handle..
int arg=0;			// For parsing command line arguments..
int DEBUG=0;			// Turn debugging statements on or off.
int gridSize=512;		// Grid side..
int pixelBuffer[512][512]={0};	// Image buffer for the PGM we've loaded..
int startBuffer[512][512]={0};	// Image buffer for the PGM we've loaded..
int hilbertIndex=0;		// Our current position on the Hilbert-Moore curve
int z=0;			// Throw-away temp variable, usually the brightness of the current pixel.
int x=0;		   	// Current X coordinate..
int y=0;		  	// Current Y coordinate..
int t=0;			// Temp swap variable..
int whiteThreshold=255;		// Pixels with brightness values above this number will be considered white.
int blackThreshold=20;		// Pixels with brightness values below this number will be considered black.
int originalPixelCount=0;	// The original number of non-black pixels of the original image.
int pixelCount=0;		// The number of pixels in the image buffer after counting.
int removalAttempts=0;		// The number of attempted removals so far.
int pixelsRemoved=0;		// The number of pixels removed to achieve the desired goal.
int targetCount=0;		// The desired number of pixels this image should be reduced to.
int thisPixel=0;		// Brightness of the current pixel.
int xPreviewSize=120;		// Width of the ASCII art preview in characters.
int yPreviewSize=65;		// Height of the ASCII art preview in characters.
int startX=0;			// The X coordinate of the salesman's starting point.
int startY=0;			// The Y coordinate of the salesman's starting point.
int pixelsWritten=0;		// Number of pixels committed to file.
int fillerPixels=0;		// The total amount of filler written to complete the frame..
char *inputFileName;		// The name of the PGM file we're going to read in..
char *outputFileName;		// The name of the PCM file we're going to write to..
char previewSize[8];		// Shows the preview size dimensions in the status line.
char c=0;			// Value of current pixel being read from file.
char gradient[]=" `.-'_,:;><+!rc*/z?LT)J(|F{CfIeZYja2SwqPhdVOKXH#$@Bg0NWQ%&M";
char header[]="Hilligoss-HMSFC-512 v1.0:";

void readImage(const char* inputFileName);
void showPreview();
void countPixels();
void reducePixels();
void determinePath();
void hilbertRotate(int n, int *x, int *y, int rotateX, int rotateY);
void getHilbertCoordinates(int n, int d, int *x, int *y);

// Allman for the people.

int main(int argc, char* argv[])
{
	if (argc<2)
	{
		printf("\nUsage: %s -f <filename> -c <desired vectors per frame, usually 2000> -b <black threshold 0-255> -w <white threshold 1-255> -x <preview width in chars> -y <preview height in chars> -s <search radius, 1-255> -j <jump frequency 1-255>\nExample: %s -f image.pgm -c2000 -b20 -w252 -x120 -y65 -j50 \nNotes: Images must be 8-bit binary PGM, 512x512 only.\n\n",argv[0],argv[0]);
	exit(0);
	}

	while ((arg=getopt(argc,argv,"f:c:b:w:x:y:j:"))!=-1)
	{
		switch(arg)
		{
			case 'f': inputFileName=optarg; break;
			case 'c': targetCount=atoi(optarg); break;
			case 'b': blackThreshold=atoi(optarg); break;
			case 'w': whiteThreshold=atoi(optarg); break;
			case 'x': xPreviewSize=atoi(optarg); break;
			case 'y': yPreviewSize=atoi(optarg); break;
		}
	}

	outputFileName=malloc(strlen(inputFileName)+1);
	strcpy(outputFileName,inputFileName);
	strcpy(outputFileName+strlen(inputFileName)-3,"pcm");

	readImage(inputFileName);
	reducePixels();
	determinePath();
	showPreview();

	free(outputFileName);
   	return 0;
}


void readImage(const char* inputFileName)
{
	DEBUG && printf("DEBUG: Inside readImage..\n");

	inputFile = fopen(inputFileName, "r");

	for(z=0;z<3;z++) while ((c=fgetc(inputFile))!='\n'); 	// We can safely ignore the header of the PGM file. 

	for (y=0;y<gridSize;y++)
	{
		for (x=0;x<gridSize;x++)
		{
			fread(&thisPixel,sizeof(unsigned char),1,inputFile); 
			if (thisPixel < blackThreshold) thisPixel=0;
			if (thisPixel > whiteThreshold) thisPixel=255;
			pixelBuffer[x][y] = thisPixel;
			startBuffer[x][y] = thisPixel;
			if (thisPixel>0) originalPixelCount++;
		}
	}

	DEBUG && printf("DEBUG: readImage: Found %d original pixels.\n",originalPixelCount);
	pixelCount=originalPixelCount;
	fclose(inputFile);
	DEBUG && printf("DEBUG: Leaving readImage..\n");
}


void showPreview(void)
{
        DEBUG && printf("DEBUG: Inside showPreview..\n");
        char output[262144]={0};
        int pos=0;

	sprintf(previewSize,"%dx%d",xPreviewSize,yPreviewSize);

        pos+=sprintf(output+pos, "\n\n%s Input:  [%15s] Source: [%6dpx] White Threshold: [%3d/255] Filler:  [%7dpx]\n                          Output: [%15s] Target: [%6dpx] Black Threshold: [%3d/255] Preview: [%7spx]\n\n",header, inputFileName, originalPixelCount, whiteThreshold, fillerPixels, outputFileName, targetCount, blackThreshold, previewSize);

        pos+=sprintf(output+pos,"p");

        for (x=0;x<=(xPreviewSize*2);x++)
        {
                pos+=sprintf(output+pos,"-");
        }

        pos+=sprintf(output+pos,"q\n");

        for (y=0;y<yPreviewSize;y++)
        {
                pos+=sprintf(output+pos,"|");

                for (x=0;x<xPreviewSize;x++)
                {
                        pos+=sprintf(output+pos,"%c",gradient[(startBuffer[x*512/xPreviewSize][y*512/yPreviewSize]*(strlen(gradient)))/256]);
                }

                pos+=sprintf(output+pos,"|");

                for (x=0;x<xPreviewSize;x++)
                {
                        pos+=sprintf(output+pos,"%c",gradient[(pixelBuffer[x*512/xPreviewSize][y*512/yPreviewSize]*(strlen(gradient)))/256]);
                }

                pos+=sprintf(output+pos,"|");

                pos+=sprintf(output+pos,"\n");
        }

        pos+=sprintf(output+pos,"b");

        for (x=0;x<=(xPreviewSize*2);x++)
        {
                pos+=sprintf(output+pos,"-");
        }

        pos+=sprintf(output+pos,"d\n");


        printf("%s\n", output);
        DEBUG && printf("DEBUG: Leaving showPreview..\n");
}

void reducePixels(void)
{
	struct timeval tv;
	gettimeofday(&tv,NULL);
	srand(tv.tv_sec^tv.tv_usec^getpid());

	if (pixelCount<=targetCount) return;

	while (pixelCount>targetCount && removalAttempts<1000000) 
	{
		x=rand()%gridSize;
		y=rand()%gridSize;
		
		if (pow(rand()%whiteThreshold,1.04)>=pixelBuffer[x][y])
		{
			if (pixelBuffer[x][y]!=0)
			{
				pixelBuffer[x][y]=0;
				pixelCount--;
				pixelsRemoved++;
				removalAttempts=0;
			}
		}

		removalAttempts++;
	}

	DEBUG && printf("DEBUG: Leaving reducePixels. %d pixels removed.\n",pixelsRemoved);
	countPixels();
}


void countPixels(void)
{
	DEBUG && printf("DEBUG: Inside countPixels..\n");

	pixelCount=0;

	for (x=0;x<gridSize;x++)
	{
		for (y=0;y<gridSize;y++)
		{
			if (pixelBuffer[x][y]>0) pixelCount++;
		}
	}
	
	DEBUG && printf("countPixels: Found %d pixels.\n",pixelCount);
}


void determinePath() 
{
	DEBUG && printf("DEBUG: Inside determinePath..\n");

	struct timeval tv;
	gettimeofday(&tv,NULL);
	srand(tv.tv_sec^tv.tv_usec^getpid());
	outputFile=fopen(outputFileName,"a");

	int16_t xShort,yShort=0;
	

	for(hilbertIndex=0;hilbertIndex<pow(gridSize,2);hilbertIndex++) // For every pixel in the image..
	{								// 
									//
		getHilbertCoordinates(gridSize,hilbertIndex,&x,&y); 	// Fetch the Hilbert coordinates of where we are along the curve
									//
		if (pixelBuffer[x][y])					// If there's a pixel there..
		{							//
			xShort=(int16_t)(x & 0x1FF);     		// Convert X and Y to 16-bit values while also 
			yShort=(gridSize/2)-(int16_t)(y & 0x1FF); 	// flipping the image along Y axis while we're at it...
									//
			xShort=xShort*128; 				// Scale up the image to encompass the full range of 16-bit values
			yShort=yShort*128; 				// else the resulting image will be kinda dinky
									//
			xShort=xShort-32767; 				// Weird offset adjustment needed for some reason..
									//	
			fwrite(&xShort,sizeof(int16_t),1,outputFile);	// Write 9-bit value as 16-bit signed int, left audio channel
			fwrite(&yShort,sizeof(int16_t),1,outputFile);	// Write 9-bit value as 16-bit signed int, right audio channel
									//
			pixelsWritten++;				//
		}							//
	}								//
									//
	while (pixelsWritten<targetCount)				// For frames which are so empty that they have fewer than the
	{								// target number of pixels, we need to sort of fill in the
		xShort=rand()%gridSize;					// gap in the sample data to prevent a short write, and prevent
		yShort=rand()%gridSize;					// batch processing from losing synchronization. Every frame needs 
									// to have the same length. Short runs will eventually add up, and
		xShort=xShort*128;					// the image will gradually fall out of sync if we dont do this.
		yShort=yShort*128;					// 
		fwrite(&xShort,sizeof(int16_t),1,outputFile); 		//
		fwrite(&yShort,sizeof(int16_t),1,outputFile); 		//
									// Thought about just painting around the perimeter of the image,
		fillerPixels++;						// but even then, there's the potential of not having enough
		pixelsWritten++;					// pixels to do that, either.
	}								//

	fclose(outputFile);

	DEBUG && printf("DEBUG: Leaving determinePath..\n");
	DEBUG && printf("\n%s Frame rendering completed. %d pixels created in absence of real data. %d stereo samples written as %d bytes to %s Exiting..\n\n", header, fillerPixels,targetCount,pixelsWritten*4,outputFileName);
	
}


void hilbertRotate(int gridSize, int *x, int *y, int rotateX, int rotateY)
{
	if (rotateY == 0)
	{
		if (rotateX == 1)
		{
			*x=gridSize-1-*x;
			*y=gridSize-1-*y;
		}

		t = *x; *x = *y; *y = t;
	}
}


void getHilbertCoordinates(int gridSize, int hilbertDistance, int *x, int *y)
{
	int rotateX,rotateY,currentScale,tempDistance=hilbertDistance;	 
      									 
	*x=*y=0;							 

	for (currentScale=1;currentScale<gridSize;currentScale*=2) 	// While crawling the Hilber curve..
	{								//
		rotateX=(tempDistance/2)%2;  				// Rotate X..
		rotateY=(tempDistance^rotateX)%2;			// Rotate Y..
		hilbertRotate(currentScale,x,y,rotateX,rotateY);	// Decide where to go..
		*x+=currentScale*rotateX;				// Get current X coordinate..
		*y+=currentScale*rotateY;				// Get current Y coordinate..
		tempDistance/=4;					// Temporary rotateY distance..
	}
}


