//
// Hilligoss-NNSTS-512 v1.1 released 012925:0228 by Bus Error Collective
// 
// Hilligoss will convert any 8-bit 512x512 Binary/"Raw" PGM image to PCM stereo audio data
// suitable for layback on XY-mode oscilloscopes and similar visualizers. It is flexible 
// enough to accommodate a wide range of sample/frame rates, and smart enough about image
// quality to render excellent looking output. 
//
// Hilligoss uses an image reduction and conversion algorithm which favors bright pixels
// over dark pixels when rendering.
//
// Nearest Neighbor Simplified Travelling Salesman (NNSTS) version

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>
#include <float.h>
#include <limits.h>
#include <string.h>

#define gridSize 512

FILE* inputFile;                // File handle..
FILE* outputFile;               // File handle..
int arg=0;                      // For parsing command line arguments..
int DEBUG=0;                    // Turn debugging statements on or off.
int visited[512][512]={0};      // Map of visited pixels..
int pixelBuffer[512][512]={0};  // Image buffer for the PGM we've loaded..
int startBuffer[512][512]={0};  // Image buffer for the PGM we've loaded..
int i=0;                        // Throw-away temp variable, usually the brightness of the current pixel.
int x=0;                        // Current X coordinate..
int y=0;                        // Current Y coordinate..
int whiteThreshold=255;         // Pixels with brightness values above this number will be considered white.
int blackThreshold=0;           // Pixels with brightness values below this number will be considered black.
int originalPixelCount=0;       // The original number of non-black pixels of the original image.
int pixelCount=0;               // The number of pixels in the image buffer after counting.
int removalAttempts=0;          // The number of attempted removals so far.
int pixelsRemoved=0;            // The number of pixels removed to achieve the desired goal.
int jumpFrequency=250;		// How often the travelling salesman randomly jumps to a new spot..
int targetCount=0;              // The desired number of pixels this image should be reduced to.
int thisPixel=0;                // Brightness of the current pixel.
int xPreviewSize=80;            // Width of the ASCII art preview in characters.
int yPreviewSize=45;            // Height of the ASCII art preview in characters.
int startX=0;                   // The X coordinate of the salesman's starting point.
int startY=0;                   // The Y coordinate of the salesman's starting point.
int pixelsWritten=0;            // Number of pixels committed to file.
int bytesWritten=0;             // The total number of bytes written to outputfilename.pcm..
int fillerPixels=0;             // The total amount of filler written to complete the frame..
int pixelList[262144][2];	// List of coordinates to help the travelling salesman..
int pixelListCount=0;		// Counter for said list..
int16_t xShort=0;		// Temporary X coordinate to be written out to file..
int16_t yShort=0;		// Temporary Y coordinate to be written out to file..
char *inputFileName;            // The name of the PGM file we're going to read in..
char *outputFileName;           // The name of the PCM file we're going to write to..
char previewSize[8];		// Preview window size string for status line..
char c=0;                       // Value of current pixel being read from file.
char gradient[]=" `.-'_,:;><+!rc*/z?LT)J(|F{CfIeZYja2SwqPhdVOKXH#$@Bg0NWQ%&M";
char header[]="Hilligoss-NNSTS-512 v1.0:";

void readImage(const char* inputFileName);
void showPreview();
void reducePixels();
void determinePath();

int main(int argc, char* argv[])
{
        if (argc<2)
        {
                printf("\nUsage: %s -f <filename> -c <desired vectors per frame, usually 2000> -b <black threshold 0-255> -w <white threshold 1-255> -x <preview width in chars> -y <preview height in chars>\nExample: %s -f image.pgm -c2000 -b20 -w234 -x160 -y100 -j200 \nNotes: Images must be 8-bit ASCII PGM, 512x512 only.\n\n",argv[0],argv[0]);
        exit(0);
        }

        while ((arg=getopt(argc,argv,"f:c:b:w:x:y:d:j:"))!=-1)
        {
                switch(arg)
                {
                        case 'f': inputFileName=optarg; break;
			case 'c': targetCount=atoi(optarg); break;
                        case 'b': blackThreshold=atoi(optarg); break;
                        case 'w': whiteThreshold=atoi(optarg); break;
                        case 'x': xPreviewSize=atoi(optarg); break;
                        case 'y': yPreviewSize=atoi(optarg); break;
			case 'j': jumpFrequency=atoi(optarg); break;
			case 'd': DEBUG=1; break;
                }
        }

        outputFileName=malloc(strlen(inputFileName)+1);
        strcpy(outputFileName,inputFileName);
        strcpy(outputFileName+strlen(inputFileName)-3,"pcm");

        readImage(inputFileName);
        reducePixels();
        determinePath();
        showPreview();

        free(outputFileName);
        return 0;
}

void readImage(const char* inputFileName)
{
        DEBUG && printf("DEBUG: Inside readImage..\n");

        inputFile = fopen(inputFileName, "r");

        for(i=0;i<3;i++) while ((c=fgetc(inputFile))!='\n');    // We can safely ignore the header of the PGM file. 

        for (y=0;y<gridSize;y++)
        {
                for (x=0;x<gridSize;x++)
                {
                        fread(&thisPixel,sizeof(unsigned char),1,inputFile);
                        if (thisPixel < blackThreshold) thisPixel=0;
                        if (thisPixel > whiteThreshold) thisPixel=255;
                        pixelBuffer[x][y] = thisPixel;
                        startBuffer[x][y] = thisPixel;
                        if (thisPixel>0) originalPixelCount++;
                }
        }

        DEBUG && printf("DEBUG: readImage: Found %d original pixels.\n",originalPixelCount);
        pixelCount=originalPixelCount;
        fclose(inputFile);
        DEBUG && printf("DEBUG: Leaving readImage..\n");
}

void showPreview(void)
{
        DEBUG && printf("DEBUG: Inside showPreview..\n");
        char output[524200+1024]={0};  // Just enough but not too far..
        int pos=0;

	sprintf(previewSize,"%dx%d",xPreviewSize,yPreviewSize);

        pos+=sprintf(output+pos, "\n\n%s Input:  [%15s] Source: [%6dpx] White Threshold: [%3d/255] Filler:  [%7dpx]\n                          Output: [%15s] Target: [%6dpx] Black Threshold: [%3d/255] Preview: [%7spx]\n\n",header, inputFileName, originalPixelCount, whiteThreshold, fillerPixels, outputFileName, targetCount, blackThreshold, previewSize);

	pos+=sprintf(output+pos,"p");

	for (x=0;x<=(xPreviewSize*2);x++)
	{
		pos+=sprintf(output+pos,"-");
	}	

	pos+=sprintf(output+pos,"q\n");

        for (y=0;y<yPreviewSize;y++)
        {
		pos+=sprintf(output+pos,"|");

                for (x=0;x<xPreviewSize;x++)
                {
                        pos+=sprintf(output+pos,"%c",gradient[(startBuffer[x*512/xPreviewSize][y*512/yPreviewSize]*(strlen(gradient)))/256]);
                }
			
		pos+=sprintf(output+pos,"|");

                for (x=0;x<xPreviewSize;x++)
                {
                        pos+=sprintf(output+pos,"%c",gradient[(pixelBuffer[x*512/xPreviewSize][y*512/yPreviewSize]*(strlen(gradient)))/256]);
		}
		
		pos+=sprintf(output+pos,"|");
                
		pos+=sprintf(output+pos,"\n");
        }
	
	pos+=sprintf(output+pos,"b");

	for (x=0;x<=(xPreviewSize*2);x++)
	{
		pos+=sprintf(output+pos,"-");
	}	

	pos+=sprintf(output+pos,"d\n");


        printf("%s\n", output);
        DEBUG && printf("DEBUG: Leaving showPreview..\n");
}

void reducePixels(void)
{
        struct timeval tv;
	int probabilityCurve=0;
	int diceRoll=0;
        gettimeofday(&tv,NULL);
        srand(tv.tv_sec^tv.tv_usec^getpid());

        if (pixelCount<=targetCount) return;

        while (pixelCount>targetCount && removalAttempts<1000000)
        {
                x=rand()%gridSize;
                y=rand()%gridSize;

                /*if (rand()%whiteThreshold>=pixelBuffer[x][y]) Hilligoss 1.0*/
		//probabilityCurve=(int)whiteThreshold*pow(cos(0.005*pixelBuffer[x][y]),9); // Secret sauce
		//printf("WT: %4d    Value: %5d   Curve: %5d   DiceRoll: %5d\n",whiteThreshold, pixelBuffer[x][y],probabilityCurve,diceRoll);


		//if (pixelBuffer[x][y] <= rand()%255)


		if (rand()%whiteThreshold<sin(3.1415926535/2*(whiteThreshold-pixelBuffer[x][y])/whiteThreshold)*whiteThreshold*0.9) // Extremely secret sauce
                {
                        if (pixelBuffer[x][y]!=0)
                        {
                                pixelBuffer[x][y]=0;
                                pixelCount--;
                                pixelsRemoved++;
                                removalAttempts=0;
                        }
                }

                removalAttempts++;
        }

        DEBUG && printf("DEBUG: Leaving reducePixels. %d pixels removed.\n",pixelsRemoved);
}


void determinePath()
{
	DEBUG && printf("DEBUG: Inside determinePath..\n");

	outputFile=fopen(outputFileName, "a");

	int pixelList[gridSize*gridSize][2];
	int pixelListCount=0;

	for (x = 0; x < gridSize; x++)
	{
		for (y = 0; y < gridSize; y++)
		{
			if (pixelBuffer[x][y] > 0)
			{
				pixelList[pixelListCount][0] = x;
				pixelList[pixelListCount][1] = y;
				pixelListCount++;
			}
		}
	}

	int path[targetCount][2];
	int visited[pixelListCount];
	for (int i=0; i<pixelListCount; i++) visited[i]=0;
	int pathLength=0;

	while (pathLength<targetCount && pathLength<pixelListCount)
	{
		int startPoint=rand()%pixelListCount;

		while (visited[startPoint]) startPoint=rand()%pixelListCount;

		path[pathLength][0] = pixelList[startPoint][0];
		path[pathLength][1] = pixelList[startPoint][1];
		visited[startPoint] = 1;
		pathLength++;

		for (int jumpCount=1; jumpCount<jumpFrequency && pathLength<targetCount; jumpCount++)
		{
			int closestIndex = -1;
			double minDistance = DBL_MAX;

			for (int i=0; i<pixelListCount; i++)
			{
				if (!visited[i])
				{
					double dist=sqrt(pow(path[pathLength-1][0]-pixelList[i][0],2) + pow(path[pathLength-1][1]-pixelList[i][1],2));
					if (dist < minDistance)
					{
						minDistance=dist;
						closestIndex=i;
					}
				}
			}

			if (closestIndex!=-1) // Add that fucker to the path
			{
				path[pathLength][0]=pixelList[closestIndex][0];
				path[pathLength][1]=pixelList[closestIndex][1];
				visited[closestIndex]=1;
				pathLength++;
			}
			else
			{
				break; // Break the loop if no unvisited pixels are found
			}
		}

		if (pathLength==pixelListCount)
		{
			break;
		}
	}

	if (pixelListCount<0.01*targetCount)
	{
		while (pixelsWritten < targetCount)
		{
			xShort = (rand() % gridSize) * 128 - 32767;
			yShort = (rand() % gridSize) * 128 - 32767;

			fwrite(&xShort, sizeof(int16_t), 1, outputFile);
			fwrite(&yShort, sizeof(int16_t), 1, outputFile);

			bytesWritten += 4;
			pixelsWritten++;
			fillerPixels++;
		}
	}
	else
	{
		for (i= 0;i<pathLength;i++)
		{
			xShort = (int16_t)(path[i][0] & 0x1FF)*128-32767;
			yShort = -(int16_t)(path[i][1] & 0x1FF)*128-32767;

			fwrite(&xShort,sizeof(int16_t),1,outputFile);
			fwrite(&yShort,sizeof(int16_t),1,outputFile);

			bytesWritten+=4;
			pixelsWritten++;
		}

		i=0;

		while (pixelsWritten < targetCount)
		{
			xShort = (int16_t)(path[i][0] & 0x1FF)*128-32767;
			yShort = -(int16_t)(path[i][1] & 0x1FF)*128-32767;

			fwrite(&xShort, sizeof(int16_t),1,outputFile);
			fwrite(&yShort, sizeof(int16_t),1,outputFile);

			bytesWritten+=4;
			fillerPixels++;
			pixelsWritten++;

			i++;

			if (i>=pathLength)
			{
				i = 0; // Loop back to start if we reach the end of the path
			}
		}
	}

	fclose(outputFile);

	DEBUG && printf("DEBUG: Leaving determinePath..\n");
	DEBUG && printf("\n%s Frame rendering completed. %d stereo samples written as %d bytes to %s. Filler points: %d. Exiting..\n\n", header, targetCount, bytesWritten, outputFileName, fillerPixels);
}




